<?php

/**
 * DocBlock for OcApi class
 *
 * Wrapper class to create an open API towards Open Content
 *
 * @author Infomaker Scandinavie AB
 *
 */

/*
|--------------------------------------------------------------------------
| Open Content API Class
| Used to interact with Open Content
|--------------------------------------------------------------------------
*/

class OcAPI extends OcConnection
{
    /**
     * @var OpenContent
     */
    private static $_oc;

    /**
     * Public construct
     * Will make a singleton Open Content object
     *
     * @scope public
     */
    public function __construct()
    {
        self::$_oc = self::$_oc ?? OpenContent::getInstance();
        parent::__construct(self::$_oc);
    }

    /**
     * Function to prepare a query
     * Will append a prepared query onto the OC URL
     *
     * @param string $query
     * @param string $start
     * @param string $limit
     * @param string $sort
     *
     * @return $this
     */
    public function prepare_query(string $query, string $start = '', string $limit = '', string $sort = ''): OcAPI
    {
        $this->oc_url = $this->prep_oc_query($query, $start, $limit, $sort);

        return $this;
    }

    /**
     * Function to prepare a Widgets oc query
     *
     * @param array $instance
     *
     * @return $this
     */
    public function prepare_widget_query(array $instance = []): OcAPI
    {
        $query = $instance['oc_query'] ?? null;
        $start = $instance['oc_query_start'] ?? null;
        $limit = $instance['oc_query_limit'] ?? null;
        $sort = $this->validateSortOption($instance['oc_query_sort'] ?? '');

        $this->oc_url = $this->prep_oc_query($query, $start, $limit, $sort);

        return $this;
    }

    /**
     * Function to prepare a suggest query
     *
     * @param array $instance
     *
     * @return void
     */
    public function prepare_suggest_query(array $instance = []): void
    {
        $suggest = $instance['oc_suggest_field'] ?? null;
        $query = $instance['oc_query'] ?? null;

        $this->prepare_oc_suggest_query($suggest, $query);
    }

    /**
     * Function used to prepare a query before its sent to Event Notifier
     *
     * @param array $instance
     *
     * @return string
     */
    public function prepare_notifier_query(array $instance = []): string
    {
        $query = $instance['oc_query'] ?? null;
        $start = $instance['oc_query_start'] ?? null;
        $limit = $instance['oc_query_limit'] ?? null;
        $sort = $instance['oc_query_sort'] ?? null;

        return $this->_build_query_string($query, $start, $limit, $sort, $url_encoded = true);
    }

    /**
     * Function to get OC Content types
     *
     * @return array
     *
     * @deprecated as of 2.1.0
     */
    public function get_content_types(): array
    {
        return $this->content_types()->getAllIndexFields();
    }

    /**
     * Fetch property names from OC, filter by contenttype and data type.
     *
     * @param string $contenttype
     * @param string $type
     *
     * @return array
     *
     * @deprecated as of 2.1.0
     */
    public function get_properties_by_type(string $contenttype, string $type): array
    {
        return $this->content_types()->getContentTypePropertiesOfType($contenttype, [$type]);
    }

    /**
     * Function to make a raw unfiltered query against Open Content
     *
     * @param String $query
     * @param string $start
     * @param string $limit
     * @param string $sort
     *
     * @return bool|null|string [JSON] Result form Open Content
     */
    public function get_raw_query($query, $start = '', $limit = '', $sort = '')
    {
        $this->oc_url = $this->prep_oc_query($query, $start, $limit, $sort);

        return $this->get_json_result($this->oc_url);
    }

    /**
     * Function to get full OC Query url data
     *
     * @param string $oc_query_url
     *
     * @return null|string
     */
    public function get_url_content($oc_query_url)
    {
        return $this->get_json_result($oc_query_url);
    }

    /**
     * Get last used query.
     *
     * @return null|string
     */
    public function get_active_query()
    {
        return $this->oc_url;
    }
}
