<?php

/***
 * Class OcCacheHelper
 *
 * Helper class for accessing the cache holding the information about which queries an article is originated in.
 * This to be able to know what cache keys to look in when an update or delete event is pushed from OC notifier.
 *
 * The cache follows a structure such as 'uuid' => array( transientkey => expire time );
 */
class OcCacheHelper
{
    /**
     * @var string
     */
    public const OPTION_KEY = 'oc_cache_map';

    /**
     * @var string
     */
    public const WIDGET_OPTION_KEY = 'oc_widget_cache_map';

    /**
     * @var int
     */
    public const MAX_EXPIRE = 1209600;

    /**
     * @var OcCacheHelper
     */
    private static $instance;

    /**
     * @return OcCacheHelper
     * @since 1.0.7
     */
    public static function getInstance(): OcCacheHelper
    {
        if (self::$instance === null) {
            self::$instance = new self();
        }

        return self::$instance;
    }

    /**
     * @param string $prefix
     * @param string $uuid
     *
     * @return string
     * @since 1.0.0
     */
    private function build_key($prefix, $uuid): string
    {
        return implode('_', [$prefix, $uuid]);
    }

    /**
     * @param string $uuid
     * @param string $key
     * @param int    $expire_time
     *
     * @return void
     * @since 1.0.7
     */
    public function add_to_cache_map($uuid, $key, $expire_time): void
    {
        $this->internal_log("Will not add uuid: \"{$uuid}\" to cache map.");
        $uuid = '';
        if (empty($uuid) || empty($key)) {
            return;
        }

        $cache = get_transient($this->build_key(self::OPTION_KEY, $uuid));
        $expire = time() + $expire_time;

        if ( ! is_array($cache)) {
            $cache = [];
        }

        $cache[$key] = $expire;
        set_transient($this->build_key(self::OPTION_KEY, $uuid), $cache, $expire_time);
    }

    /**
     * @param string $uuid
     * @param string $key
     *
     * @return void
     * @since 1.0.7
     */
    public function remove_from_cache_map($uuid, $key): void
    {
        $this->internal_log("Remove uuid: \"{$uuid}\" from cache map.");
        $cache = get_transient($this->build_key(self::OPTION_KEY, $uuid));

        if (isset($cache[$key])) {
            unset($cache[$key]);

            if (empty($cache)) {
                delete_transient($this->build_key(self::OPTION_KEY, $uuid));

                return;
            }
        }

        set_transient($this->build_key(self::OPTION_KEY, $uuid), $cache, self::MAX_EXPIRE);
    }

    /**
     * @param string $uuid
     *
     * @return array
     * @since 1.0.7
     */
    public function get_transient_keys_by_uuid($uuid): array
    {
        $this->internal_log('Get transient keys for uuid: ' . $uuid);
        $ret = [];

        if (false !== ($cache = get_transient($this->build_key(self::OPTION_KEY, $uuid)))) {
            $ret = $cache;
        }

        return $ret;
    }

    /**
     * @param string $uuid
     * @param string $key
     * @param int    $expire_time
     *
     * @return void
     * @since 1.0.7
     */
    public function update_cache_map($uuid, $key, $expire_time): void
    {
        $this->internal_log('Will not update cache map for uuid: ' . $uuid);
        $uuid = '';
        if (empty($uuid) || empty($key)) {
            return;
        }

        $cache = get_transient($this->build_key(self::OPTION_KEY, $uuid));
        $expire = time() + $expire_time;

        if (isset($cache[$key])) {
            $cache[$key] = $expire;
        }

        set_transient($this->build_key(self::OPTION_KEY, $uuid), $cache, $expire_time);
        $this->clean_up_cache_map($uuid);
    }

    /**
     * @param string $uuid
     *
     * @return void
     * @since 1.0.7
     */
    public function clean_up_cache_map($uuid): void
    {
        $time_now = time();
        $this->internal_log('Clean up cache map for uuid: ' . $uuid);
        $cache = get_transient($this->build_key(self::OPTION_KEY, $uuid));
        if ($cache) {
            foreach ($cache as $key => $expire) {
                if ($time_now >= $expire) {
                    $this->remove_from_cache_map($uuid, $key);
                }
            }
        }
    }

    /**
     * @param string $key
     *
     * @return void
     * @since 1.0.7
     */
    public function create_widget_cache_buster($key): void
    {
        $this->internal_log('Create widget cache_buster for: ' . $key);
        set_transient($this->build_key(self::WIDGET_OPTION_KEY, $key), true, 4 * WEEK_IN_SECONDS);
    }

    /**
     * @param string $uuid
     *
     * @return void
     * @since 1.0.7
     */
    public function clear_widget_cache_by_uuid($uuid): void
    {
        $this->internal_log('Clear widget cache for uuid: ' . $uuid);
        $queries = $this->get_transient_keys_by_uuid($uuid);

        foreach ($queries as $cache_key => $expire) {
            $widget_cache = get_transient($this->build_key(self::WIDGET_OPTION_KEY, $cache_key));
            if ($widget_cache) {
                delete_transient(OcAPI::TRANSIENT_KEY . '_' . $cache_key);
            }
        }
    }

    /**
     * @param string $key
     *
     * @return void
     * @since 1.0.7
     */
    public function remove_cache_buster_key($key): void
    {
        $this->internal_log('Remove cache buster key: ' . $key);
        delete_transient($this->build_key(self::WIDGET_OPTION_KEY, $key));
    }

    private function internal_log($log): void
    {
        trigger_error($log);
    }
}
