<?php declare(strict_types=1);

namespace Everyware\Helpers;

use Everyware\Exceptions\ResponseCacheError;
use Everyware\Storage\OcResponseCache;
use Everyware\Storage\Traits\MemoryStorage;
use GuzzleHttp\Exception\GuzzleException;
use JsonException;

class OcContentTypesProvider
{
    use MemoryStorage;

    /**
     * @var string
     */
    public const MEMORY_KEY = 'oc_content_types';

    /**
     * @var OcResponseCache
     */
    private $cache;

    public function __construct(OcResponseCache $cache)
    {
        $this->cache = $cache;
    }

    public function contentTypeIndexFieldExists(string $contentType, string $indexField): bool
    {
        return ! empty(
        array_filter(
            $this->getContentTypeIndexFields($contentType),
            static function ($i) use ($indexField) {
                return in_array($indexField, $i, true);
            })
        );
    }

    public function contentTypePropertyExists(string $contentType, string $property): bool
    {
        return in_array($property, $this->getContentTypeProperties($contentType), true);
    }

    public function getAllIndexFields(): array
    {
        return $this->getFilteredIndexFields();
    }

    public function getAllIndexFieldsOfType(array $types): array
    {
        return $this->getFilteredIndexFields([], $types);
    }

    public function getAllProperties(): array
    {
        return $this->getFilteredProperties();
    }

    public function getAllPropertiesOfType(array $types): array
    {
        return $this->getFilteredProperties([], $types);
    }

    public function getAllRelatedProperties(): array
    {
        return $this->getFilteredProperties([], $this->getContentTypes());
    }

    public function getContentTypeIndexFields(string $contentType): array
    {
        return $this->getFilteredIndexFields((array)$contentType)[$contentType] ?? [];
    }

    public function getContentTypeIndexFieldsOfType(string $contentType, array $types): array
    {
        return $this->getFilteredIndexFields((array)$contentType, $types)[$contentType] ?? [];
    }

    public function getContentTypeProperties(string $contentType): array
    {
        return $this->getFilteredProperties((array)$contentType)[$contentType] ?? [];
    }

    public function getContentTypePropertiesOfType(string $contentType, array $types): array
    {
        return $this->getFilteredProperties((array)$contentType, $types)[$contentType] ?? [];
    }

    public function getContentTypeRelatedProperties(string $contentType): array
    {
        return $this->getFilteredProperties([], $this->getContentTypes())[$contentType] ?? [];
    }

    public function getContentTypes(): array
    {
        return array_keys($this->getStoredContentTypes());
    }

    public function getFilteredIndexFields(array $contentTypes = [], array $types = []): array
    {
        $indexFields = array_map(function ($props) use ($types) {
            return $this->filterIndexFieldsByType($props, $types);
        }, $this->getStoredContentTypesByName($contentTypes));

        return empty($contentTypes) ? array_filter($indexFields) : $indexFields;
    }

    public function getFilteredProperties(array $contentTypes = [], array $types = []): array
    {
        $properties = array_map(function ($props) use ($types) {
            return $this->filterPropertiesByType($props, $types);
        }, $this->getStoredContentTypesByName($contentTypes));

        return empty($contentTypes) ? array_filter($properties) : $properties;
    }

    public function getRawJson(): string
    {
        return $this->getRawJsonResponse();
    }

    protected function filterPropertiesByType(array $properties, array $types = []): array
    {
        return $this->extractPropertyValues($this->extractPropertyTypes($properties, $types), 'name');
    }

    protected function filterIndexFieldsByType(array $properties, array $types): array
    {
        return $this->extractPropertyValues($this->extractPropertyTypes($properties, $types), 'indexFields');
    }

    protected function getRawJsonResponse(): string
    {
        try {
            if ($this->inMemory(static::MEMORY_KEY)) {
                return $this->getFromMemory(static::MEMORY_KEY);
            }

            $response = $this->cache->getContent(
                'contenttypes',
                [],
                new ContentTypesResponseValidator()
            );

            $this->addToMemory(static::MEMORY_KEY, $response);

            return $response;
        } catch (GuzzleException|ResponseCacheError $e) {
            NewRelicLogger::logException($e);
        }

        return '';
    }

    protected function getStoredContentTypes(): array
    {
        $response = $this->getRawJsonResponse();

        if (empty($response)) {
            return [];
        }

        try {
            $responseJson = json_decode($response, true);

            $contentTypes = $responseJson['contentTypes'] ?? [];

            if ( ! empty($contentTypes)) {
                return array_column($contentTypes, 'properties', 'name');
            }

        } catch (\Exception $e) {
            NewRelicLogger::logException($e);
        }

        return [];
    }

    protected function getStoredContentTypesByName(array $names = []): array
    {
        $contentTypes = $this->getStoredContentTypes();

        $useFilter = ! empty($names);

        $return = [];
        foreach ($contentTypes as $contentType => $properties) {
            if ( ! $useFilter || in_array($contentType, $names, true)) {
                $return[$contentType] = $properties;
            }
        }

        return $return;
    }

    protected function extractPropertyTypes(array $properties, array $types = []): array
    {
        if (empty($types)) {
            return $properties;
        }

        $types = array_map('strtolower', $types);

        return array_values(array_filter($properties, static function ($property) use ($types) {
            return in_array(strtolower($property['type']), $types, true);
        }));
    }

    protected function extractPropertyValues(array $properties, string $valueName): array
    {
        $return = [];

        foreach ($properties as $property) {
            $value = $property[$valueName];

            if ( ! empty($value)) {
                $return[] = $value;
            }
        }

        return $return;
    }
}
