<?php declare(strict_types=1);

namespace Everyware;

use Everyware\Contracts\OcObject;
use Everyware\Exceptions\ObjectNotFoundException;
use Everyware\Exceptions\OcConnectionException;
use Everyware\Exceptions\PersistentOcConnectionException;
use Everyware\OcObjects\ContentType;
use Everyware\OcObjects\EmptyOcObject;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ConnectException;
use GuzzleHttp\Exception\GuzzleException;
use OpenContent;
use Psr\Http\Message\ResponseInterface;

class OcClient
{
    public const OC_VERSION_PATH = 'infoandstats/version';

    public const OC_HEALTH_PATH = 'health';

    public const RESPONSE_TIMEOUT = 10;

    public const CONNECTION_TIMEOUT = 10;

    public const MAX_CONNECTION_FAILURES = 2;

    private static $connectionFailCount = 0;

    /**
     * @var Client
     */
    private $client;

    public function __construct(Client $client)
    {
        $this->client = $client;
    }

    /**
     * @param string $uri
     * @param array  $options
     *
     * @return ResponseInterface
     * @throws GuzzleException
     */
    public function get(string $uri, array $options = []): ResponseInterface
    {
        $this->verifyOcConnection('GET', $uri);

        try {
            $response = $this->client->get($uri, $this->withTimeoutOptions($options));
        } catch (ConnectException $exception) {
            $this->saveConnectionFailure();
            throw new OcConnectionException($exception);
        }

        return $response;
    }

    /**
     * @param int|null $id
     * @param array    $params
     *
     * @return array
     * @throws GuzzleException
     */
    public function events(int $id = null, array $params = []): array
    {
        // Make sure to add the event as parameter
        if (is_int($id)) {
            $params['event'] = $id;
        }

        $response = $this->getContent('eventlog', $params);
        $body = json_decode($response, true);

        return $body['events'] ?? [];
    }

    /**
     * @param string $uri
     * @param array  $params
     *
     * @return string|null
     * @throws GuzzleException
     */
    public function getContent(string $uri, array $params = []): string
    {
        $options = [];

        if ( ! empty($params)) {
            $options['query'] = $params;
        }

        $response = $this->get($uri, $options);

        return $response->getBody()->getContents() ?? '';
    }

    /**
     * @param string $uuid
     * @param array  $params
     *
     * @return OcObject
     * @throws GuzzleException
     */
    public function getObject(string $uuid, array $params = []): OcObject
    {
        try {
            $content = $this->getContent("objects/$uuid/properties", $params);

            return $this->createOcObject($uuid, $content);
        } catch (GuzzleException $e) {
            $this->reValidateGuzzleException($e);
        }
    }

    /**
     * @return string|null
     * @throws GuzzleException
     */
    public function getOcVersion(): ?string
    {
        return $this->getContent(static::OC_VERSION_PATH);
    }

    /**
     * @param string $query
     * @param array  $params
     *
     * @return array
     * @throws GuzzleException
     */
    public function search(string $query, array $params = []): array
    {
        // Make sure to add the query as parameter
        if ( ! empty($query)) {
            $params['q'] = $query;
        }

        $response = $this->getContent('search', $params);
        $body = json_decode($response, true);

        return $body['hits'] ?? [];
    }

    /**
     * @return bool
     * @throws GuzzleException
     */
    public function testConnection(): bool
    {
        return $this->testPath(static::OC_VERSION_PATH, 200);
    }

    /**
     * @param string $user
     * @param string $password
     *
     * @return bool
     * @throws GuzzleException
     */
    public function testCredentials(string $user, string $password): bool
    {
        return $this->testPath(static::OC_VERSION_PATH, 200, [
            'auth' => [$user, $password]
        ]);
    }

    /**
     * @return bool
     * @throws GuzzleException
     */
    public function testHealth(): bool
    {
        return $this->testPath(static::OC_HEALTH_PATH, 200);
    }

    /**
     * @param string $path
     * @param int    $code
     * @param array  $options
     *
     * @return bool
     * @throws GuzzleException
     */
    public function testPath(string $path, int $code, array $options = []): bool
    {
        $this->verifyOcConnection('HEAD', $path);

        $options['http_errors'] = false;
        $response = $this->client->request('head', $path, $options);

        return $response->getStatusCode() === $code;
    }

    /**
     * @param OpenContent $ocInstance
     *
     * @return OcClient
     */
    public static function createFromWpSettings(OpenContent $ocInstance): OcClient
    {
        return static::create(
            $ocInstance->getOcBaseUrl() ?? '',
            $ocInstance->getOcUserName(),
            $ocInstance->getOcPassword()
        );
    }

    /**
     * @param $uri
     * @param $user
     * @param $password
     *
     * @return OcClient
     */
    public static function create($uri, $user, $password): OcClient
    {
        $clientOptions = [
            'base_uri' => $uri,
            'auth' => [$user, $password],
            'headers' => [
                'Content-Type' => 'application/x-www-form-urlencoded; charset=utf-8',
                'Accept' => 'text/plain, application/json;'
            ]
        ];

        return new static(new Client($clientOptions));
    }

    /**
     * @param string $uuid
     * @param string $content
     *
     * @return OcObject
     */
    private function createOcObject(string $uuid, string $content): OcObject
    {
        return ! empty($content) ? new ContentType($content) : new EmptyOcObject($uuid, $content);
    }

    /**
     * @param GuzzleException $exception
     *
     * @throws GuzzleException
     */
    private function reValidateGuzzleException(GuzzleException $exception): void
    {
        if ( ! $exception instanceof ClientException) {
            throw $exception;
        }

        $notFoundCode = [403, 404, 410];

        if (in_array((int)$exception->getCode(), $notFoundCode)) {
            throw new ObjectNotFoundException($exception);
        }

        throw $exception;
    }

    private function withTimeoutOptions(array $options): array
    {
        return $options + [
                'timeout' => self::RESPONSE_TIMEOUT,
                'connect_timeout' => self::CONNECTION_TIMEOUT
            ];
    }

    private function hasConnectionFailure(): bool
    {
        return static::$connectionFailCount >= static::MAX_CONNECTION_FAILURES;
    }

    private function saveConnectionFailure(): void
    {
        static::$connectionFailCount++;
    }

    /**
     * @throws OcConnectionException
     */
    private function verifyOcConnection(string $method, string $uri): void
    {
        if ($this->hasConnectionFailure()) {
            throw new PersistentOcConnectionException($method, $uri);
        }
    }

    public function resetConnectionFailureState(): void
    {
        static::$connectionFailCount = 0;
    }
}
