<?php declare(strict_types=1);

namespace Everyware\Storage;

use Everyware\Contracts\OcResponseCacheValidator;
use Everyware\Exceptions\InvalidOcResponseException;
use Everyware\Exceptions\OcConnectionException;
use Everyware\Exceptions\PersistentOcConnectionException;
use Everyware\Exceptions\ResponseCacheError;
use Everyware\Helpers\NewRelicLogger;
use Everyware\OcClient;
use Everyware\Storage\Contracts\SimpleCacheInterface;
use Exception;
use GuzzleHttp\Exception\BadResponseException;
use GuzzleHttp\Exception\ConnectException;
use GuzzleHttp\Exception\GuzzleException;
use Psr\SimpleCache\InvalidArgumentException;

class OcResponseCache
{
    /**
     * @var string
     */
    public const KEY_PREFIX = 'oc_r_';

    /**
     * TTL is 1h in seconds
     * @var int
     */
    public const DEFAULT_MAX_TTL = 3600;

    /**
     * @var OcClient
     */
    private $client;

    /**
     * @var SimpleCacheInterface
     */
    private $cache;

    /**
     * @var int
     */
    private $ttl;

    public function __construct(OcClient $client, SimpleCacheInterface $cache, int $ttl = 0)
    {
        $this->client = $client;
        $this->cache = $cache;
        $this->ttl = abs($ttl);
    }

    /**
     * @throws GuzzleException
     * @throws ResponseCacheError
     */
    public function getContent(string $uri, array $params = [], OcResponseCacheValidator $validator = null): string
    {
        try {
            $key = $this->generateKey($uri, $params);
            $cacheObject = $this->cache->get($key);
        } catch (InvalidArgumentException $e) {
            throw new ResponseCacheError(sprintf('Failed to create proper cache key out of uri: "%s"', $uri));
        }

        // Generate new cache if no cache or if the cached data is not instance of CacheObject e.g. has an old cache structure
        if ( ! $cacheObject instanceof CacheObject) {
            $cacheObject = $this->generateCache($uri, $params, $validator);
        }

        if ($this->shouldUpdateCache($cacheObject)) {
            try {
                $cacheObject = $this->generateCache($uri, $params, $validator);
            } catch (GuzzleException $e) {

                if ($this->shouldRemoveCache($e)) {
                    $this->removeStoredResponseObject($key);
                    throw $e;
                }

                $this->logException($e);
                $this->resetStoredResponseObject($key, $cacheObject);

            } catch (ResponseCacheError $e) {
                $this->logException($e);
                $this->resetStoredResponseObject($key, $cacheObject);
            }
        }

        return $cacheObject->getValue();
    }

    /**
     * @throws GuzzleException
     * @throws ResponseCacheError
     */
    private function generateCache(
        string $uri,
        array $params = [],
        OcResponseCacheValidator $validator = null
    ): CacheObject {
        $key = $this->generateKey($uri, $params);

        try {
            $response = $this->client->getContent($uri, $params);

            if ($validator instanceof OcResponseCacheValidator) {
                $validator->validateResponse($response);
            }

        } catch (InvalidOcResponseException|OcConnectionException $e) {
            throw new ResponseCacheError($e->getMessage());
        }

        $cacheObject = new CacheObject($response);

        if ( ! $this->storeResponseObject($key, $cacheObject)) {
            throw new ResponseCacheError(sprintf('Could not add response of request against "%s" to cache', $uri));
        }

        return $cacheObject;
    }

    public function getMaxCacheTTL()
    {
        if (defined('PHP_RESPONSE_CACHE_TTL')) {
            return PHP_RESPONSE_CACHE_TTL;
        }

        return static::DEFAULT_MAX_TTL;
    }

    public function generateKey(string $uri, array $params = []): string
    {
        return self::KEY_PREFIX . md5($uri . http_build_query($params));
    }

    protected function storeResponseObject(string $key, CacheObject $cacheObject): bool
    {
        try {
            return $this->cache->set($key, $cacheObject, $this->getMaxCacheTTL());
        } catch (InvalidArgumentException $e) {
            $this->logException($e);

            return false;
        }
    }

    protected function resetStoredResponseObject(string $key, CacheObject $cacheObject): bool
    {
        $cacheObject->resetAge();

        return $this->storeResponseObject($key, $cacheObject);
    }

    protected function removeStoredResponseObject(string $key): bool
    {
        try {
            return $this->cache->delete($key);
        } catch (InvalidArgumentException $e) {
            $this->logException($e);

            return false;
        }
    }

    protected function shouldUpdateCache(CacheObject $cacheObject): bool
    {
        return $cacheObject->ageExceeds($this->ttl);
    }

    protected function logException(Exception $e): void
    {
        NewRelicLogger::logException($e);
    }

    protected function shouldRemoveCache(GuzzleException $e): bool
    {
        $codeBlackList = [403, 404, 410];

        return in_array($e->getCode(), $codeBlackList);
    }
}
