<?php declare(strict_types=1);

namespace Spec\Everyware\Helpers;

use Everyware\Helpers\ContentTypesResponseValidator;
use Everyware\Helpers\OcContentTypesProvider;
use Everyware\Storage\OcResponseCache;
use PhpSpec\ObjectBehavior;
use Prophecy\Argument;
use Spec\DocumentFileLoader;

class OcContentTypesProviderSpec extends ObjectBehavior
{

    /**
     * @var OcResponseCache
     */
    private $cache;

    public function let(OcResponseCache $cache): void
    {
        $this->cache = $cache;

        $this->beConstructedWith($cache);
    }

    public function letGo(): void
    {
        $this->clearMemory();
    }

    public function it_is_initializable(): void
    {
        $this->shouldHaveType(OcContentTypesProvider::class);
    }

    public function it_should_store_and_return_the_raw_json_format_of_the_open_content_response(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesResponseJson());

        $this->getRawJson()->shouldReturn($this->contentTypesResponseJson());
    }

    public function it_should_fetch_all_available_content_types_from_an_open_content(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesResponseJson());

        $this->getContentTypes()->shouldReturn([
            'Article',
            'Concept',
            'Image',
            'List',
            'Package'
        ]);
    }

    public function it_should_fetch_all_index_fields_grouped_by_contenttype(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getAllIndexFields()->shouldReturn([
            'Article' => [
                ['checksumIndexField'],
                ['contenttypeIndexField'],
                ['createdIndexField']
            ],
            'Concept' => [
                ['DescriptionIndexField'],
                ['updatedIndexField'],
                ['uuidIndexField']
            ]
        ]);
    }

    public function it_should_ignore_empty_index_fields(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJsonWithMissingIndexFields());

        $this->getAllIndexFields()->shouldReturn([
            'Article' => [
                ['contenttypeIndexField'],
                ['createdIndexField']
            ],
            'Concept' => [
                ['DescriptionIndexField'],
                ['uuidIndexField']
            ]
        ]);
    }

    public function it_should_fetch_all_index_fields_of_type_from_specific_contenttype(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getAllIndexFieldsOfType(['date'])->shouldReturn([
            'Article' => [
                ['createdIndexField']
            ],
            'Concept' => [
                ['updatedIndexField']
            ]
        ]);
    }

    public function it_should_ignore_empty_index_fields_of_the_specified_type_from_specific_contenttype(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJsonWithMissingIndexFields());

        $this->getAllIndexFieldsOfType(['date'])->shouldReturn([
            'Article' => [
                ['createdIndexField']
            ]
        ]);
    }

    public function it_should_fetch_all_properties_grouped_by_contenttype(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getAllProperties()->shouldReturn([
            'Article' => [
                'checksum',
                'contenttype',
                'created'
            ],
            'Concept' => [
                'Description',
                'updated',
                'uuid'
            ]
        ]);
    }

    public function it_should_fetch_all_properties_of_type_from_specific_contenttype(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getAllPropertiesOfType(['date'])->shouldReturn([
            'Article' => [
                'created'
            ],
            'Concept' => [
                'updated'
            ]
        ]);
    }

    public function it_should_fetch_index_fields_based_on_a_specific_contenttype(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getContentTypeIndexFields('Concept')->shouldReturn([
            ['DescriptionIndexField'],
            ['updatedIndexField'],
            ['uuidIndexField']
        ]);
    }

    public function it_should_fetch_index_fields_of_type_based_on_a_specific_contenttype(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getContentTypeIndexFieldsOfType('Concept', ['date'])->shouldReturn([
            ['updatedIndexField']
        ]);
    }

    public function it_should_fetch_properties_based_on_a_specific_contenttype(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getContentTypeProperties('Article')->shouldReturn([
            'checksum',
            'contenttype',
            'created'
        ]);
    }

    public function it_should_fetch_properties_of_type_based_on_a_specific_contenttype(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getContentTypePropertiesOfType('Article', ['date'])->shouldReturn([
            'created'
        ]);
    }

    public function it_should_fetch_index_fields_based_on_multiple_contenttypes(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getFilteredIndexFields(['Article', 'Concept'])->shouldReturn([
            'Article' => [
                ['checksumIndexField'],
                ['contenttypeIndexField'],
                ['createdIndexField']
            ],
            'Concept' => [
                ['DescriptionIndexField'],
                ['updatedIndexField'],
                ['uuidIndexField']
            ]
        ]);
    }

    public function it_should_not_include_empty_index_fields_for_multiple_contenttypes(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJsonWithMissingIndexFields());

        $this->getFilteredIndexFields(['Article', 'Concept'])->shouldReturn([
            'Article' => [
                ['contenttypeIndexField'],
                ['createdIndexField']
            ],
            'Concept' => [
                ['DescriptionIndexField'],
                ['uuidIndexField']
            ]
        ]);
    }

    public function it_should_ignore_empty_content_types_if_contenttype_is_not_specified(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJsonWithMissingIndexFields());

        $this->getFilteredIndexFields([], ['date'])->shouldReturn([
            'Article' => [
                ['createdIndexField']
            ]
        ]);
    }

    public function it_should_fetch_properties_based_on_multiple_contenttypes(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getFilteredProperties(['Article', 'Concept'])->shouldReturn([
            'Article' => [
                'checksum',
                'contenttype',
                'created'
            ],
            'Concept' => [
                'Description',
                'updated',
                'uuid'
            ]
        ]);
    }

    public function it_should_not_include_unknown_content_types_while_filtering_index_fields_based_on_contenttypes(
    ): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getFilteredIndexFields(['Article', 'Concept', 'UnknownType'])->shouldReturn([
            'Article' => [
                ['checksumIndexField'],
                ['contenttypeIndexField'],
                ['createdIndexField']
            ],
            'Concept' => [
                ['DescriptionIndexField'],
                ['updatedIndexField'],
                ['uuidIndexField']
            ]
        ]);
    }

    public function it_should_not_include_unknown_content_types_while_filtering_properties_based_on_contenttypes(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesSmallResponseJson());

        $this->getFilteredProperties(['Article', 'Concept', 'UnknownType'])->shouldReturn([
            'Article' => [
                'checksum',
                'contenttype',
                'created'
            ],
            'Concept' => [
                'Description',
                'updated',
                'uuid'
            ]
        ]);
    }

    public function it_should_find_related_properties_from_a_content_type(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesResponseJson());

        $this->getContentTypeRelatedProperties('Article')->shouldReturn([
            'Articles',
            'Authors',
            'Categories',
            'Concepts',
            'Places',
            'Sections',
            'Stories',
            'Tags',
        ]);
    }

    public function it_can_determine_with_case_sensitivity_if_a_property_exists_on_a_specific_content_type(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesResponseJson());

        $this->contentTypePropertyExists('Article', 'uuid')->shouldReturn(true);
        $this->contentTypePropertyExists('Article', 'Uuid')->shouldReturn(false);
        $this->contentTypePropertyExists('Article', 'unknown')->shouldReturn(false);
    }

    public function it_can_determine_with_case_sensitivity_if_an_indexfield_exists_on_a_specific_content_type(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesResponseJson());

        $this->contentTypeIndexFieldExists('Article', 'uuid')->shouldReturn(true);
        $this->contentTypeIndexFieldExists('Article', 'Uuid')->shouldReturn(false);
        $this->contentTypeIndexFieldExists('Article', 'unknown')->shouldReturn(false);
    }

    public function it_should_find_all_contenttypes_with_related_properties_grouped_by_their_names(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesResponseJson());

        $this->getAllRelatedProperties()->shouldReturn([
            'Article' => [
                'Articles',
                'Authors',
                'Categories',
                'Concepts',
                'Places',
                'Sections',
                'Stories',
                'Tags',
            ],
            'Concept' => [
                'Articles',
                'Parent',
            ],
            'List' => [
                'Articles',
                'Lists',
                'Packages',
                'Subjects',
            ],
            'Package' => [
                'CoverArticle',
                'PackageList',
            ],

        ]);
    }

    public function it_should_be_able_to_filter_properties_both_by_contenttype_and_property_type(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesResponseJson());

        $this->getFilteredProperties(['Article'], ['date'])->shouldReturn([
            'Article' => [
                'created',
                'indexed_date',
                'updated',
                'ObjectUpdated',
                'Pubdate'
            ],
        ]);
    }

    public function it_should_return_filtered_properties_for_all_contenttypes_if_not_provided(): void
    {
        $this->simulateGetContentTypesFromSource($this->contentTypesResponseJson());

        $this->getFilteredProperties([], ['date'])->shouldReturn([
            'Article' => [
                'created',
                'indexed_date',
                'updated',
                'ObjectUpdated',
                'Pubdate'
            ],
            'Concept' => [
                'created',
                'indexed_date',
                'updated',
            ],
            'Image' => [
                'created',
                'indexed_date',
                'updated',
            ],
            'List' => [
                'created',
                'indexed_date',
                'updated',
            ],
            'Package' => [
                'created',
                'indexed_date',
                'updated',
                'PubStart',
                'PubStop',
            ],
        ]);
    }

    public function it_should_store_fetched_properties_in_memory(): void
    {
        $this->cache->getContent(
            'contenttypes',
            [],
            Argument::type(ContentTypesResponseValidator::class)
        )->shouldBeCalledTimes(1)->willReturn($this->contentTypesResponseJson());

        $this->getFilteredProperties(['Article'], ['date'])->shouldReturn([
            'Article' => [
                'created',
                'indexed_date',
                'updated',
                'ObjectUpdated',
                'Pubdate'
            ],
        ]);

        $this->getFilteredProperties(['Article'], ['integer'])->shouldReturn([
            'Article' => [
                'version',
                'EditionTeaserImageHeight',
                'EditionTeaserImageWidth',
                'NewsScore',
                'TeaserImageHeight',
                'TeaserImageWidth'
            ],
        ]);
    }

    private function simulateGetContentTypesFromSource(string $response): void
    {
        $this->cache->getContent(
            'contenttypes',
            [],
            Argument::type(ContentTypesResponseValidator::class)
        )->willReturn($response);
    }

    private function contentTypesResponseJson(): string
    {
        return DocumentFileLoader::loadDocumentJson('full_contenttypes_response.json');
    }

    private function contentTypesSmallResponseJson(): string
    {
        return <<<JSON
{
    "contentTypes": [
        {
            "name": "Article",
            "properties": [
                {
                    "name": "checksum",
                    "type": "STRING",
                    "multiValued": false,
                    "searchable": false,
                    "readOnly": true,
                    "description": "Concatenated md5 checksums of all the files that are part of the object.",
                    "indexFields": [
                        "checksumIndexField"
                    ]
                },
                {
                    "name": "contenttype",
                    "type": "STRING",
                    "multiValued": false,
                    "searchable": false,
                    "readOnly": true,
                    "description": "contenttype specifies the type of object.",
                    "indexFields": [
                        "contenttypeIndexField"
                    ]
                },
                {
                    "name": "created",
                    "type": "DATE",
                    "multiValued": false,
                    "searchable": false,
                    "readOnly": true,
                    "description": "created is the date and time at which the object was created in Open Content.",
                    "indexFields": [
                        "createdIndexField"
                    ]
                }
            ]
        },
        {
            "name": "Concept",
            "properties": [
                {
                    "name": "Description",
                    "type": "STRING",
                    "multiValued": false,
                    "searchable": true,
                    "readOnly": false,
                    "description": "DO NOT DELETE OR CHANGE... Full description of the Concept extracted on long defenition from newsML",
                    "indexFields": [
                        "DescriptionIndexField"
                    ]
                },
                {
                    "name": "updated",
                    "type": "DATE",
                    "multiValued": false,
                    "searchable": false,
                    "readOnly": true,
                    "description": "updated is the date and time at which the object was updated in Open Content.",
                    "indexFields": [
                        "updatedIndexField"
                    ]
                },
                {
                    "name": "uuid",
                    "type": "STRING",
                    "multiValued": false,
                    "searchable": false,
                    "readOnly": true,
                    "description": "uuid identifies the object. It is preferably a UUID, but it can be any string.",
                    "indexFields": [
                        "uuidIndexField"
                    ]
                }
            ]
        }
    ]
}
JSON;
    }

    private function contentTypesSmallResponseJsonWithMissingIndexFields(): string
    {
        return <<<JSON
{
    "contentTypes": [
        {
            "name": "Article",
            "properties": [
                {
                    "name": "checksum",
                    "type": "STRING",
                    "multiValued": false,
                    "searchable": false,
                    "readOnly": true,
                    "description": "Concatenated md5 checksums of all the files that are part of the object.",
                    "indexFields": []
                },
                {
                    "name": "contenttype",
                    "type": "STRING",
                    "multiValued": false,
                    "searchable": false,
                    "readOnly": true,
                    "description": "contenttype specifies the type of object.",
                    "indexFields": [
                        "contenttypeIndexField"
                    ]
                },
                {
                    "name": "created",
                    "type": "DATE",
                    "multiValued": false,
                    "searchable": false,
                    "readOnly": true,
                    "description": "created is the date and time at which the object was created in Open Content.",
                    "indexFields": [
                        "createdIndexField"
                    ]
                }
            ]
        },
        {
            "name": "Concept",
            "properties": [
                {
                    "name": "Description",
                    "type": "STRING",
                    "multiValued": false,
                    "searchable": true,
                    "readOnly": false,
                    "description": "DO NOT DELETE OR CHANGE... Full description of the Concept extracted on long defenition from newsML",
                    "indexFields": [
                        "DescriptionIndexField"
                    ]
                },
                {
                    "name": "updated",
                    "type": "DATE",
                    "multiValued": false,
                    "searchable": false,
                    "readOnly": true,
                    "description": "updated is the date and time at which the object was updated in Open Content.",
                    "indexFields": []
                },
                {
                    "name": "uuid",
                    "type": "STRING",
                    "multiValued": false,
                    "searchable": false,
                    "readOnly": true,
                    "description": "uuid identifies the object. It is preferably a UUID, but it can be any string.",
                    "indexFields": [
                        "uuidIndexField"
                    ]
                }
            ]
        }
    ]
}
JSON;
    }

    private function createStoredValue(string $responseJson)
    {
        $storedContentTypes = json_decode($responseJson, true);

        return $storedContentTypes['contentTypes'];
    }
}
