<?php
/**
 * @noinspection PhpUndefinedMethodInspection
 * @noinspection PhpUnhandledExceptionInspection
 * @noinspection ReturnTypeCanBeDeclaredInspection
 * @noinspection AccessModifierPresentedInspection
 * @noinspection PhpUnused
 */

namespace Spec\Everyware;

use Everyware\Contracts\OcObject;
use Everyware\Exceptions\ObjectNotFoundException;
use Everyware\Exceptions\OcConnectionException;
use Everyware\Exceptions\PersistentOcConnectionException;
use Everyware\OcClient;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ConnectException;
use GuzzleHttp\Psr7\Request;
use GuzzleHttp\Psr7\Response;
use PhpSpec\ObjectBehavior;
use Prophecy\Argument;
use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\StreamInterface;

class OcClientSpec extends ObjectBehavior
{
    /**
     * @var Client
     */
    private $client;

    /**
     * @var ResponseInterface
     */
    private $response;

    /**
     * @var StreamInterface
     */
    private $stream;

    function let(Client $client, ResponseInterface $response, StreamInterface $stream)
    {
        $this->client = $client;
        $this->response = $response;
        $this->stream = $stream;

        $this->beConstructedWith($client);
    }

    function letGo()
    {
        $this->resetConnectionFailureState();
    }

    function it_is_initializable()
    {
        $this->shouldHaveType(OcClient::class);
    }

    function it_should_support_general_get_requests()
    {
        $uri = 'test';

        $this->client->get($uri, Argument::type('array'))->willReturn($this->response);

        $this->get($uri)->shouldReturn($this->response);
    }

    function it_should_add_timeout_options_to_general_get_requests()
    {
        $uri = 'test';

        $this->client->get($uri, $this->combineStandardClientOptions([]))->shouldBeCalled();

        $this->get($uri, []);
    }

    function it_should_support_a_way_to_override_the_timeout_options_for_general_get_requests()
    {
        $uri = 'test';
        $newResponseTimeout = OcClient::RESPONSE_TIMEOUT * 2;
        $newConnectionTimeout = OcClient::CONNECTION_TIMEOUT * 2;

        $this->client->get($uri, [
            'other_option' => true,
            'timeout' => $newResponseTimeout,
            'connect_timeout' => $newConnectionTimeout
        ])->shouldBeCalled();

        $this->get($uri, [
            'other_option' => true,
            'timeout' => $newResponseTimeout,
            'connect_timeout' => $newConnectionTimeout
        ]);
    }

    function it_should_support_general_get_requests_with_content_as_response()
    {
        $uri = 'test';
        $content = json_encode(['testing']);
        $this->client->get($uri, Argument::type('array'))->willReturn($this->response);

        $this->simulateResponseContent($content);

        $this->getContent($uri)->shouldReturn($content);
    }

    function it_should_support_search_requests()
    {
        $query = 'test';
        $params = ['q' => $query];

        $content = ['content'];

        $this->client->get(
            'search',
            $this->combineStandardClientOptions(['query' => $params])
        )->willReturn($this->response);

        $this->simulateResponseJson(['hits' => $content]);

        $this->search($query, $params)->shouldReturn($content);
    }

    function it_should_support_search_requests_with_additional_parameters()
    {
        $query = 'test';
        $addedParams = [
            'contenttype' => 'Article',
            'properties' => 'uuid'
        ];

        $params = array_replace($addedParams, [
            'q' => $query
        ]);

        $content = ['content'];

        $this->client->get(
            'search',
            $this->combineStandardClientOptions(['query' => $params])
        )->willReturn($this->response);

        $this->simulateResponseJson(['hits' => $content]);

        $this->search($query, $addedParams)->shouldReturn($content);
    }

    function it_should_support_requests_for_events()
    {
        $id = 1;

        $content = ['content'];

        $this->client->get(
            'eventlog',
            $this->combineStandardClientOptions(['query' => ['event' => $id]])
        )->willReturn($this->response);

        $this->simulateResponseJson(['events' => $content]);

        $this->events($id)->shouldReturn($content);
    }

    function it_should_not_send_event_id_unless_given_while_requesting_events()
    {
        $content = ['content'];

        $this->client->get('eventlog', Argument::type('array'))->willReturn($this->response);

        $this->simulateResponseJson(['events' => $content]);

        $this->events()->shouldReturn($content);
    }

    function it_should_be_able_to_retrieve_document_by_id()
    {
        $content = json_encode(['content']);
        $uuid = '000000-0000-0000-0000';

        $this->client->get("objects/$uuid/properties", Argument::type('array'))->willReturn($this->response);

        $this->simulateResponseJson($content);

        $this->getObject($uuid)->shouldBeAnInstanceOf(OcObject::class);
    }

    function it_can_handle_404_status_code()
    {
        $this->assertConversionToObjectNotFoundException(404);
    }

    function it_will_handle_403_status_code_as_NOT_FOUND()
    {
        $this->assertConversionToObjectNotFoundException(403);
    }

    function it_will_handle_410_status_code_as_NOT_FOUND()
    {
        $this->assertConversionToObjectNotFoundException(410);
    }

    function it_can_determine_if_path_will_return_200_status_code()
    {
        $path = 'path';
        $code = 200;
        $this->client->request('head', $path, ['http_errors' => false])->willReturn($this->response);

        $this->response->getStatusCode()->willReturn($code);

        $this->testPath($path, $code)->shouldReturn(true);
        $this->testPath($path, 500)->shouldReturn(false);
    }

    function it_can_determine_if_options_to_path_will_return_error_codes()
    {
        $path = 'path';
        $code = 400;
        $options = ['query' => ['foo' => 'bar']];
        $this->client->request('head', $path, array_replace($options, ['http_errors' => false]))
            ->willReturn($this->response);

        $this->response->getStatusCode()->willReturn($code);

        $this->testPath($path, $code, $options)->shouldReturn(true);
        $this->testPath($path, 200, $options)->shouldReturn(false);
    }

    function it_can_determine_the_open_content_version()
    {
        $version = '2.4.0';
        $this->client->get(OcClient::OC_VERSION_PATH, Argument::type('array'))->willReturn($this->response);

        $this->simulateResponseContent($version);

        $this->getOcVersion()->shouldReturn($version);
    }

    function it_can_test_the_connection_to_open_content()
    {
        $this->client->request('head', OcClient::OC_VERSION_PATH,
            ['http_errors' => false])->willReturn($this->response);
        $this->response->getStatusCode()->willReturn(200);

        $this->testConnection()->shouldReturn(true);

        $this->response->getStatusCode()->willReturn(500);

        $this->testConnection()->shouldReturn(false);
    }

    function it_can_test_credentials_against_an_open_content()
    {
        $user = 'user';
        $password = 'pass';
        $this->client->request('head', OcClient::OC_VERSION_PATH, [
            'auth' => [$user, $password],
            'http_errors' => false
        ])->willReturn($this->response);

        $this->response->getStatusCode()->willReturn(200);

        $this->testCredentials($user, $password)->shouldReturn(true);

        $this->response->getStatusCode()->willReturn(500);

        $this->testCredentials($user, $password)->shouldReturn(false);
    }

    function it_can_test_the_health_of_open_content()
    {
        $this->client->request('head', OcClient::OC_HEALTH_PATH, ['http_errors' => false])->willReturn($this->response);
        $this->response->getStatusCode()->willReturn(200);

        $this->testHealth()->shouldReturn(true);

        $this->response->getStatusCode()->willReturn(500);

        $this->testHealth()->shouldReturn(false);
    }

    function it_should_convert_ConnectExceptions_to_our_own_Exception_type()
    {
        $this->client->get(Argument::type('string'), Argument::type('array'))
            ->willThrow($this->generateConnectionException());

        $this->shouldThrow(OcConnectionException::class)->duringGet(Argument::type('string'));
    }

    function it_should_enter_stale_mode_on_persistent_connection_issues_with_Open_Content()
    {
        $connectException = $this->generateConnectionException();

        $this->client->get(Argument::type('string'), Argument::type('array'))
            ->willThrow($connectException)
            ->shouldBeCalledTimes(OcClient::MAX_CONNECTION_FAILURES);

        $this->shouldThrow(OcConnectionException::class)->duringGet(Argument::type('string'));
        $this->shouldThrow(OcConnectionException::class)->duringGet(Argument::type('string'));
        $this->shouldThrow(PersistentOcConnectionException::class)->duringGet(Argument::type('string'));
        $this->shouldThrow(PersistentOcConnectionException::class)->duringGet(Argument::type('string'));
    }

    function it_has_a_manual_reset_of_the_connection_failure_state()
    {
        $connectException = $this->generateConnectionException();

        $this->client->get(Argument::type('string'), Argument::type('array'))
            ->willThrow($connectException)
            ->shouldBeCalledTimes(OcClient::MAX_CONNECTION_FAILURES * 2);

        $this->shouldThrow(OcConnectionException::class)->duringGet(Argument::type('string'));
        $this->shouldThrow(OcConnectionException::class)->duringGet(Argument::type('string'));
        $this->shouldThrow(PersistentOcConnectionException::class)->duringGet(Argument::type('string'));
        $this->shouldThrow(PersistentOcConnectionException::class)->duringGet(Argument::type('string'));

        $this->resetConnectionFailureState();

        $this->shouldThrow(OcConnectionException::class)->duringGet(Argument::type('string'));
        $this->shouldThrow(OcConnectionException::class)->duringGet(Argument::type('string'));
        $this->shouldThrow(PersistentOcConnectionException::class)->duringGet(Argument::type('string'));
        $this->shouldThrow(PersistentOcConnectionException::class)->duringGet(Argument::type('string'));
    }

    private function simulateResponseJson($content)
    {
        $this->simulateResponseContent(json_encode($content));
    }

    private function simulateResponseContent($content)
    {
        $this->response->getBody()->willReturn($this->stream);

        $this->stream->getContents()->willReturn($content);
    }

    private function generateClientException(int $code): ClientException
    {
        return new ClientException(
            'Generated ClientException',
            new Request('get', 'Generated request path for ClientException'),
            new Response($code)
        );
    }

    private function generateConnectionException(): ConnectException
    {
        return new ConnectException(
            'Generated ConnectException',
            new Request('get', 'Generated request path for ConnectException')
        );
    }

    private function assertConversionToObjectNotFoundException(int $code)
    {
        $this->client->get(
            Argument::type('string'),
            Argument::type('array')
        )->willThrow($this->generateClientException($code));

        $this->shouldThrow(ObjectNotFoundException::class)->during('getObject', [Argument::type('string')]);
    }

    private function combineStandardClientOptions(array $options): array
    {
        return [
                'timeout' => OcClient::RESPONSE_TIMEOUT,
                'connect_timeout' => OcClient::CONNECTION_TIMEOUT,
            ] + $options;
    }
}
