<?php

/** @noinspection PhpUnused */

/** @noinspection ReturnTypeCanBeDeclaredInspection */

namespace Spec\Everyware\Storage;

use Everyware\Storage\CacheObject;
use PhpSpec\ObjectBehavior;

/**
 * @method ageExceeds(int $seconds)
 * @method decreaseAge(int $seconds)
 * @method getAgeInSeconds()
 * @method getTimestamp()
 * @method getValue()
 * @method increaseAge(int $seconds)
 * @method resetAge()
 * @method toArray()
 */
class CacheObjectSpec extends ObjectBehavior
{
    /**
     * @var mixed
     */
    private $value = 'CachedValue';

    public function let()
    {
        $this->beConstructedWith($this->value);
    }

    public function it_is_initializable()
    {
        $this->shouldHaveType(CacheObject::class);
    }

    public function it_will_store_the_value()
    {
        $this->getValue()->shouldReturn($this->value);
    }

    public function it_will_store_a_timestamp_of_when_the_object_was_created()
    {
        $this->getTimestamp()->shouldReturn(time());
    }

    public function it_can_create_an_array_representation_of_itself()
    {
        $this->toArray()->shouldReturn([
            'timestamp' => time(),
            'value' => $this->value
        ]);
    }

    public function it_can_manually_increase_the_age_by_decreasing_the_timestamp()
    {
        $ageDiff = 20;
        $age = time();

        $this->increaseAge($ageDiff);
        $this->getTimestamp()->shouldReturn($age - $ageDiff);
    }

    public function it_can_manually_decrease_the_age_by_increasing_the_timestamp()
    {
        $ageDiff = 20;

        $age = time();

        $this->decreaseAge($ageDiff);
        $this->getTimestamp()->shouldReturn($age + $ageDiff);
    }

    public function it_can_manually_reset_the_age()
    {
        $this->getAgeInSeconds()->shouldReturn(0);

        $this->increaseAge(20);
        $this->getAgeInSeconds()->shouldReturn(20);

        $this->resetAge();
        $this->getAgeInSeconds()->shouldReturn(0);
    }

    public function it_will_throw_exception_when_trying_to_increase_or_decrease_age_with_negative_number()
    {
        $this->shouldThrow(\InvalidArgumentException::class)->duringIncreaseAge(-1);
        $this->shouldThrow(\InvalidArgumentException::class)->duringDecreaseAge(-1);
    }

    public function it_will_throw_exception_when_trying_to_increase_or_decrease_age_with_zero()
    {
        $this->shouldThrow(\InvalidArgumentException::class)->duringIncreaseAge(0);
        $this->shouldThrow(\InvalidArgumentException::class)->duringDecreaseAge(0);
    }

    public function it_will_return_the_age_of_the_object_in_seconds()
    {
        $this->getAgeInSeconds()->shouldReturn(0);

        $this->increaseAge(10);
        $this->getAgeInSeconds()->shouldReturn(10);

        $this->decreaseAge(5);
        $this->getAgeInSeconds()->shouldReturn(5);
    }

    public function it_can_reduce_the_age_to_a_time_before_current_time()
    {
        $this->getAgeInSeconds()->shouldReturn(0);

        $this->decreaseAge(5);
        $this->getAgeInSeconds()->shouldReturn(-5);
    }

    public function it_can_determine_if_the_age_exceed_a_certain_time_aka_has_expired()
    {
        $this->getAgeInSeconds()->shouldReturn(0);
        $this->ageExceeds(5)->shouldReturn(false);

        $this->increaseAge(10);
        $this->ageExceeds(5)->shouldReturn(true);
    }
}
