<?php

/**
 * @noinspection PhpUndefinedMethodInspection
 * @noinspection PhpUnhandledExceptionInspection
 * @noinspection PhpVoidFunctionResultUsedInspection
 * @noinspection ReturnTypeCanBeDeclaredInspection
 * @noinspection PhpUnused
 */

namespace Spec\Everyware\Storage;

use Everyware\Contracts\OcResponseCacheValidator;
use Everyware\Exceptions\InvalidCacheKey;
use Everyware\Exceptions\InvalidOcResponseException;
use Everyware\Exceptions\OcConnectionException;
use Everyware\Exceptions\ResponseCacheError;
use Everyware\OcClient;
use Everyware\Storage\CacheObject;
use Everyware\Storage\Contracts\SimpleCacheInterface;
use Everyware\Storage\OcResponseCache;
use GuzzleHttp\Exception\BadResponseException;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ConnectException;
use GuzzleHttp\Exception\ServerException;
use GuzzleHttp\Psr7\Request;
use GuzzleHttp\Psr7\Response;
use PhpSpec\ObjectBehavior;
use Prophecy\Argument;

/**
 * @method generateKey(string $uri, array $params = [])
 * @method getContent(string $uri, array $params = [])
 */
class OcResponseCacheSpec extends ObjectBehavior
{
    /**
     * @var SimpleCacheInterface
     */
    private $cache;

    /**
     * @var CacheObject
     */
    private $cacheObject;

    /**
     * @var OcClient
     */
    private $client;

    private $ttl = 60;

    private $defaultResponse = 'Response';

    public function let(OcClient $client, SimpleCacheInterface $cache, CacheObject $cacheObject)
    {
        $this->client = $client;
        $this->cache = $cache;
        $this->cacheObject = $cacheObject;


        $this->beConstructedWith($client, $cache, $this->ttl);
    }

    public function it_is_initializable()
    {
        $this->shouldHaveType(OcResponseCache::class);
    }

    public function it_should_combine_uri_and_parameters_to_generate_the_cache_key()
    {
        $uri = 'uri/to/request';

        $params = [
            'q' => 'query',
            'param' => 1
        ];

        $key = OcResponseCache::KEY_PREFIX . md5($uri . http_build_query($params));

        $this->generateKey($uri, $params)->shouldReturn($key);
    }

    public function it_should_not_include_empty_parameter_array_when_generating_key()
    {
        $uri = 'uri/to/request';
        $key = OcResponseCache::KEY_PREFIX . md5($uri);

        $this->generateKey($uri)->shouldReturn($key);
        $this->generateKey($uri, [])->shouldReturn($key);
    }

    public function it_should_store_the_result_of_the_request_in_cache()
    {
        $uri = 'uri';
        $parameters = ['parameter'];

        $this->cache->get(Argument::type('string'))->shouldBeCalled()->willReturn(null);

        $this->client->getContent($uri, $parameters)->willReturn($this->defaultResponse);

        $this->cache->set(Argument::type('string'), Argument::type(CacheObject::class), PHP_RESPONSE_CACHE_TTL)
            ->willReturn(true);

        $this->getContent($uri, $parameters)->shouldReturn($this->defaultResponse);
    }

    public function it_should_throw_error_if_a_cache_key_can_not_be_created_for_response_cache()
    {
        $this->cache->get(Argument::type('string'))->willThrow(InvalidCacheKey::class);

        $this->client->getContent(Argument::type('string'), [])->shouldNotBeCalled();

        $this->shouldThrow(ResponseCacheError::class)->duringGetContent(Argument::type('string'));
    }

    public function it_should_get_the_result_of_the_request_from_the_cache()
    {
        $this->cacheObject->getValue()->willReturn($this->defaultResponse);
        $this->cacheObject
            ->ageExceeds($this->ttl)
            ->shouldBeCalled()
            ->willReturn(false);

        $this->cache->get(Argument::type('string'))->willReturn($this->cacheObject);

        $this->getContent(Argument::type('string'))->shouldReturn($this->defaultResponse);
    }

    public function it_should_validate_the_cache_timestamp_and_update_old_cache()
    {
        $this->cacheObject->getValue()->willReturn($this->defaultResponse);
        $this->cacheObject->ageExceeds($this->ttl)->willReturn(true);

        $this->cache->get(Argument::type('string'))->willReturn($this->cacheObject);

        $this->client->getContent(Argument::type('string'), [])->willReturn($this->defaultResponse);

        $this->cache->set(Argument::type('string'), Argument::type(CacheObject::class),
            PHP_RESPONSE_CACHE_TTL)->willReturn(true);

        $this->getContent(Argument::type('string'))->shouldReturn($this->defaultResponse);
    }

    public function it_should_generate_new_cache_if_old_structure_is_encountered()
    {
        $this->cache->get(Argument::type('string'))->willReturn('Not a CacheObject');

        $this->client->getContent(Argument::type('string'), [])->willReturn($this->defaultResponse);

        $this->cache->set(Argument::type('string'), Argument::type(CacheObject::class),
            PHP_RESPONSE_CACHE_TTL)
            ->willReturn(true);

        $this->getContent(Argument::type('string'))->shouldReturn($this->defaultResponse);
    }

    public function it_should_throw_error_on_client_error_while_initially_fetching_content()
    {
        $this->cache->get(Argument::type('string'))->shouldBeCalled();

        $this->client->getContent(Argument::type('string'), [])->willThrow(ClientException::class);

        $this->cache->set(Argument::type('string'), Argument::type('string'))->shouldNotBeCalled();

        $this->shouldThrow(ClientException::class)->duringGetContent(Argument::type('string'));
    }

    public function it_should_throw_error_if_response_could_not_be_stored_in_cache()
    {
        $this->cache->get(Argument::type('string'))->shouldBeCalled();

        $this->client->getContent(Argument::type('string'), [])->willReturn($this->defaultResponse);

        $this->cache->set(Argument::type('string'), Argument::type(CacheObject::class),
            PHP_RESPONSE_CACHE_TTL)->willReturn(false);

        $this->shouldThrow(ResponseCacheError::class)->duringGetContent(Argument::type('string'));
    }

    public function it_should_remove_cache_and_throw_errors_on_403_client_errors_while_updating_the_cached_response()
    {
        $this->removeCacheAndThrowErrorDuringUpdate(403);
    }

    public function it_should_remove_cache_and_throw_errors_on_404_client_errors_while_updating_the_cached_response()
    {
        $this->removeCacheAndThrowErrorDuringUpdate(404);
    }

    public function it_should_remove_cache_and_throw_errors_on_410_client_errors_while_updating_the_cached_response()
    {
        $this->removeCacheAndThrowErrorDuringUpdate(410);
    }

    public function it_should_reset_and_return_stale_cache_on_Bad_Request_errors_while_updating_the_cached_response()
    {
        $exception = $this->generateClientException(400);

        $this->cacheObject->getValue()->willReturn($this->defaultResponse);
        $this->cacheObject->ageExceeds($this->ttl)
            ->shouldBeCalled()
            ->willReturn(true);

        $this->cacheObject->resetAge()->shouldBeCalled();

        $this->cache->get(Argument::type('string'))->willReturn($this->cacheObject);

        $this->client->getContent(Argument::type('string'), [])->willThrow($exception);

        $this->cache->set(Argument::type('string'), $this->cacheObject, PHP_RESPONSE_CACHE_TTL)
            ->shouldBeCalled()
            ->willReturn(true);

        $this->getContent(Argument::type('string'))->shouldReturn($this->defaultResponse);
    }

    public function it_should_reset_and_return_stale_cache_on_server_error_while_updating_the_cached_response()
    {
        $this->cacheObject->getValue()->willReturn($this->defaultResponse);
        $this->cacheObject->ageExceeds($this->ttl)
            ->shouldBeCalled()
            ->willReturn(true);

        $this->cacheObject->resetAge()->shouldBeCalled();

        $this->cache->get(Argument::type('string'))->willReturn($this->cacheObject);

        $this->client->getContent(Argument::type('string'), [])->willThrow(ServerException::class);

        $this->cache->set(Argument::type('string'), $this->cacheObject, PHP_RESPONSE_CACHE_TTL)
            ->shouldBeCalled()
            ->willReturn(true);

        $this->getContent(Argument::type('string'))->shouldReturn($this->defaultResponse);
    }

    public function it_should_always_reset_stale_cache_TTL_on_OC_connection_error_while_updating_cached_response()
    {
        $this->cacheObject->getValue()->willReturn($this->defaultResponse);
        $this->cacheObject->ageExceeds($this->ttl)
            ->shouldBeCalled()
            ->willReturn(true);

        $this->cacheObject->resetAge()->shouldBeCalled();

        $this->cache->get(Argument::type('string'))->willReturn($this->cacheObject);

        $this->client->getContent(Argument::type('string'), [])->willThrow(OcConnectionException::class);

        $this->cache->set(Argument::type('string'), $this->cacheObject, PHP_RESPONSE_CACHE_TTL)
            ->shouldBeCalled()
            ->willReturn(true);

        $this->getContent(Argument::type('string'))->shouldReturn($this->defaultResponse);
    }

    public function it_can_validate_responses_with_an_interface_before_adding_to_cache(
        OcResponseCacheValidator $validator
    ) {
        $validator->validateResponse($this->defaultResponse)->shouldBeCalled();

        $this->cache->get(Argument::type('string'))->shouldBeCalled();

        $this->client->getContent(Argument::type('string'), [])->willReturn($this->defaultResponse);

        $this->cache->set(Argument::type('string'), Argument::type(CacheObject::class), PHP_RESPONSE_CACHE_TTL)
            ->willReturn(true);

        $this->getContent(Argument::type('string'), [], $validator)->shouldReturn($this->defaultResponse);
    }

    public function it_should_throw_error_if_validation_fails_while_initially_fetching_content(
        OcResponseCacheValidator $validator
    ) {
        $validator->validateResponse($this->defaultResponse)->shouldBeCalled()->willThrow(InvalidOcResponseException::class);

        $this->cache->get(Argument::type('string'))->shouldBeCalled();

        $this->client->getContent(Argument::type('string'), [])->willReturn($this->defaultResponse);

        $this->cache->set(Argument::type('string'), Argument::type(CacheObject::class), PHP_RESPONSE_CACHE_TTL)
            ->shouldNotBeCalled();

        $this->shouldThrow(ResponseCacheError::class)->duringGetContent(Argument::type('string'), [], $validator);
    }

    public function it_should_return_stale_cache_on_validation_error_while_updating_the_cached_response(
        OcResponseCacheValidator $validator
    ) {
        $newResponse = 'New Response';
        $validator->validateResponse($newResponse)->shouldBeCalled()->willThrow(InvalidOcResponseException::class);

        $this->cacheObject->getValue()->willReturn($this->defaultResponse);
        $this->cacheObject->ageExceeds($this->ttl)
            ->shouldBeCalled()
            ->willReturn(true);
        $this->cacheObject->resetAge()->shouldBeCalled();

        $this->cache->get(Argument::type('string'))->willReturn($this->cacheObject);

        $this->client->getContent(Argument::type('string'), [])->willReturn($newResponse);

        $this->cache->set(Argument::type('string'), $this->cacheObject, PHP_RESPONSE_CACHE_TTL)
            ->shouldBeCalled()
            ->willReturn(true);

        $this->getContent(Argument::type('string'), [], $validator)->shouldReturn($this->defaultResponse);
    }

    private function generateClientException(int $code, string $method = 'get'): ClientException
    {
        return new ClientException(
            'Generated ClientException',
            new Request($method, 'Generated request path for ClientException'),
            new Response($code)
        );
    }

    private function removeCacheAndThrowErrorDuringUpdate($code)
    {
        $pathToRemove = 'path_to_remove';
        $cacheToRemove = $this->generateKey($pathToRemove);

        $exception = $this->generateClientException($code);

        $this->cacheObject->getValue()->willReturn($this->defaultResponse);
        $this->cacheObject->ageExceeds($this->ttl)
            ->shouldBeCalled()
            ->willReturn(true);

        $this->cacheObject->ageExceeds(3600)->shouldNotBeCalled();

        $this->cache->get($cacheToRemove)->willReturn($this->cacheObject);

        $this->client->getContent($pathToRemove, [])->willThrow($exception);

        $this->cache->delete($cacheToRemove)->shouldBeCalled()->willReturn(true);

        $this->shouldThrow($exception)->duringGetContent($pathToRemove);
    }
}
